@@ -72,7 +72,7 @@ struct Proxy {
 
 struct Mapping {
   gchar *protocol;
-  guint external_port;
+  guint requested_external_port;
   gchar *local_ip;
   guint16 local_port;
   guint32 lease_duration;
@@ -86,6 +86,7 @@ struct ProxyMapping {
   GUPnPServiceProxyAction *action;
 
   gboolean mapped;
+  guint actual_external_port;
 
   GSource *renew_src;
 };
@@ -194,7 +195,7 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
    * @external_ip: the external IP
    * @replaces_external_ip: if this mapping replaces another mapping,
    *  this is the old external IP
-   * @external_port: the external port
+   * @external_port: the external port that was allocated
    * @local_ip: internal ip this is forwarded to
    * @local_port: the local port
    * @description: the user's selected description
@@ -218,7 +219,9 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
    * @self: #GUPnPSimpleIgd that emitted the signal
    * @error: a #GError
    * @proto: The requested protocol
-   * @external_port: the requested external port
+   * @external_port: the external port requested in gupnp_simple_igd_add_port()
+   * @local_ip: internal ip this is forwarded to
+   * @local_port: the local port
    * @description: the passed description
    *
    * This means that mapping a port on a specific IGD has failed (it may still
@@ -230,9 +233,9 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
       0,
       NULL,
       NULL,
-      _gupnp_simple_igd_marshal_VOID__POINTER_STRING_UINT_STRING,
-      G_TYPE_NONE, 4, G_TYPE_POINTER, G_TYPE_STRING, G_TYPE_UINT,
-      G_TYPE_STRING);
+      _gupnp_simple_igd_marshal_VOID__POINTER_STRING_UINT_STRING_UINT_STRING,
+      G_TYPE_NONE, 6, G_TYPE_POINTER, G_TYPE_STRING, G_TYPE_UINT,
+      G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING);
 }
 
 static void
@@ -309,6 +312,11 @@ _external_ip_address_changed (GUPnPServiceProxy *proxy, const gchar *variable,
 
   g_return_if_fail (G_VALUE_HOLDS_STRING(value));
 
+  /* It hasn't really changed, ignore it */
+  if (prox->external_ip &&
+      !strcmp (g_value_get_string (value), prox->external_ip))
+    return;
+
   new_ip = g_value_dup_string (value);
 
   for (i=0; i < prox->proxymappings->len; i++)
@@ -318,7 +326,7 @@ _external_ip_address_changed (GUPnPServiceProxy *proxy, const gchar *variable,
     if (pm->mapped)
       g_signal_emit (prox->parent, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
           pm->mapping->protocol, new_ip, prox->external_ip,
-          pm->mapping->external_port, pm->mapping->local_ip,
+          pm->actual_external_port, pm->mapping->local_ip,
           pm->mapping->local_port, pm->mapping->description);
   }
 
@@ -555,22 +563,28 @@ _service_proxy_got_external_ip_address (GUPnPServiceProxy *proxy,
   g_return_if_fail (prox->external_ip_action == action);
 
   prox->external_ip_action = NULL;
-  
+
   if (gupnp_service_proxy_end_action (proxy, action, &error,
           "NewExternalIPAddress", G_TYPE_STRING, &ip,
           NULL))
   {
     guint i;
 
-    for (i=0; i < prox->proxymappings->len; i++)
-    {
-      struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
 
-      if (pm->mapped)
-        g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
-            pm->mapping->protocol, ip, prox->external_ip,
-            pm->mapping->external_port, pm->mapping->local_ip,
-            pm->mapping->local_port, pm->mapping->description);
+    /* Only emit the new signal if the IP changes */
+    if (prox->external_ip &&
+        strcmp (ip, prox->external_ip))
+    {
+      for (i=0; i < prox->proxymappings->len; i++)
+      {
+        struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
+
+        if (pm->mapped)
+          g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
+              pm->mapping->protocol, ip, prox->external_ip,
+              pm->actual_external_port, pm->mapping->local_ip,
+              pm->mapping->local_port, pm->mapping->description);
+      }
     }
 
     g_free (prox->external_ip);
@@ -588,7 +602,8 @@ _service_proxy_got_external_ip_address (GUPnPServiceProxy *proxy,
       struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
 
       g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
-          error, pm->mapping->protocol, pm->mapping->external_port,
+          error, pm->mapping->protocol, pm->mapping->requested_external_port,
+          pm->mapping->local_ip, pm->mapping->local_port,
           pm->mapping->description);
     }
   }
@@ -618,27 +633,33 @@ _service_proxy_renewed_port_mapping (GUPnPServiceProxy *proxy,
   GUPnPSimpleIgd *self = pm->proxy->parent;
   GError *error = NULL;
 
+  g_return_if_fail (pm->action == action);
+
+  pm->action = NULL;
+
   if (!gupnp_service_proxy_end_action (proxy, action, &error,
           NULL))
   {
     g_return_if_fail (error);
     g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
-        error, pm->mapping->protocol, pm->mapping->external_port,
+        error, pm->mapping->protocol, pm->mapping->requested_external_port,
+        pm->mapping->local_ip, pm->mapping->local_port,
         pm->mapping->description);
   }
   g_clear_error (&error);
 }
 
-static gboolean
-_renew_mapping_timeout (gpointer user_data)
+static void
+gupnp_simple_igd_call_add_port_mapping (struct ProxyMapping *pm,
+    GUPnPServiceProxyActionCallback callback)
 {
-  struct ProxyMapping *pm = user_data;
+  g_return_if_fail (pm->action == NULL);
 
-  gupnp_service_proxy_begin_action (pm->proxy->proxy,
+  pm->action = gupnp_service_proxy_begin_action (pm->proxy->proxy,
       "AddPortMapping",
-      _service_proxy_renewed_port_mapping, pm,
+      callback, pm,
       "NewRemoteHost", G_TYPE_STRING, "",
-      "NewExternalPort", G_TYPE_UINT, pm->mapping->external_port,
+      "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
       "NewProtocol", G_TYPE_STRING, pm->mapping->protocol,
       "NewInternalPort", G_TYPE_UINT, pm->mapping->local_port,
       "NewInternalClient", G_TYPE_STRING, pm->mapping->local_ip,
@@ -646,6 +667,17 @@ _renew_mapping_timeout (gpointer user_data)
       "NewPortMappingDescription", G_TYPE_STRING, pm->mapping->description,
       "NewLeaseDuration", G_TYPE_UINT, pm->mapping->lease_duration,
       NULL);
+}
+
+static gboolean
+_renew_mapping_timeout (gpointer user_data)
+{
+  struct ProxyMapping *pm = user_data;
+
+  stop_proxymapping (pm);
+
+  gupnp_simple_igd_call_add_port_mapping (pm,
+      _service_proxy_renewed_port_mapping);
 
   return TRUE;
 }
@@ -671,7 +703,7 @@ _service_proxy_added_port_mapping (GUPnPServiceProxy *proxy,
     if (pm->proxy->external_ip)
       g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
           pm->mapping->protocol, pm->proxy->external_ip, NULL,
-          pm->mapping->external_port, pm->mapping->local_ip,
+          pm->actual_external_port, pm->mapping->local_ip,
           pm->mapping->local_port, pm->mapping->description);
 
 
@@ -686,13 +718,26 @@ _service_proxy_added_port_mapping (GUPnPServiceProxy *proxy,
   else
   {
     g_return_if_fail (error);
-    g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
-        error, pm->mapping->protocol, pm->mapping->external_port,
-        pm->mapping->description);
+
+    /* 718 == ConflictInMappingEntry */
+    if (pm->mapping->requested_external_port == 0 &&
+        error->domain == GUPNP_CONTROL_ERROR && error->code == 718)
+    {
+      /* The previous port was already used, lets pick another random port */
+      pm->actual_external_port = g_random_int_range (1025, 65535);
+
+      gupnp_simple_igd_call_add_port_mapping (pm,
+          _service_proxy_added_port_mapping);
+    }
+    else
+    {
+      g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
+          error, pm->mapping->protocol, pm->mapping->requested_external_port,
+          pm->mapping->local_ip, pm->mapping->local_port,
+          pm->mapping->description);
+    }
   }
   g_clear_error (&error);
-
-  stop_proxymapping (pm);
 }
 
 static void
@@ -704,18 +749,13 @@ gupnp_simple_igd_add_proxy_mapping (GUPnPSimpleIgd *self, struct Proxy *prox,
   pm->proxy = prox;
   pm->mapping = mapping;
 
-  pm->action = gupnp_service_proxy_begin_action (prox->proxy,
-      "AddPortMapping",
-      _service_proxy_added_port_mapping, pm,
-      "NewRemoteHost", G_TYPE_STRING, "",
-      "NewExternalPort", G_TYPE_UINT, mapping->external_port,
-      "NewProtocol", G_TYPE_STRING, mapping->protocol,
-      "NewInternalPort", G_TYPE_UINT, mapping->local_port,
-      "NewInternalClient", G_TYPE_STRING, mapping->local_ip,
-      "NewEnabled", G_TYPE_BOOLEAN, TRUE,
-      "NewPortMappingDescription", G_TYPE_STRING, mapping->description,
-      "NewLeaseDuration", G_TYPE_UINT, mapping->lease_duration,
-      NULL);
+  if (mapping->requested_external_port)
+    pm->actual_external_port = mapping->requested_external_port;
+  else
+    pm->actual_external_port = mapping->local_port;
+
+  gupnp_simple_igd_call_add_port_mapping (pm,
+      _service_proxy_added_port_mapping);
 
   g_ptr_array_add (prox->proxymappings, pm);
 }
@@ -736,7 +776,7 @@ gupnp_simple_igd_add_port_real (GUPnPSimpleIgd *self,
   g_return_if_fail (!strcmp (protocol, "UDP") || !strcmp (protocol, "TCP"));
 
   mapping->protocol = g_strdup (protocol);
-  mapping->external_port = external_port;
+  mapping->requested_external_port = external_port;
   mapping->local_ip = g_strdup (local_ip);
   mapping->local_port = local_port;
   mapping->lease_duration = lease_duration;
@@ -758,7 +798,8 @@ gupnp_simple_igd_add_port_real (GUPnPSimpleIgd *self,
                       "Could not get external address"};
       g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT],
           GUPNP_SIMPLE_IGD_ERROR,
-          &error, mapping->protocol, mapping->external_port,
+          &error, mapping->protocol, mapping->requested_external_port,
+          mapping->local_ip, mapping->local_port,
           mapping->description);
     }
     else
@@ -772,7 +813,9 @@ gupnp_simple_igd_add_port_real (GUPnPSimpleIgd *self,
  * gupnp_simple_igd_add_port:
  * @self: The #GUPnPSimpleIgd object
  * @protocol: the protocol "UDP" or "TCP"
- * @external_port: The port to try to open on the external device
+ * @external_port: The port to try to open on the external device,
+ *   0 means to try a random port if the same port as the local port is already
+ *   taken
  * @local_ip: The IP address to forward packets to (most likely the local ip address)
  * @local_port: The local port to forward packets to
  * @lease_duration: The duration of the lease (it will be auto-renewed before it expires). This is in seconds.
@@ -836,7 +879,7 @@ gupnp_simple_igd_remove_port_real (GUPnPSimpleIgd *self,
   for (i = 0; i < self->priv->mappings->len; i++)
   {
     struct Mapping *tmpmapping = g_ptr_array_index (self->priv->mappings, i);
-    if (tmpmapping->external_port == external_port &&
+    if (tmpmapping->requested_external_port == external_port &&
         !strcmp (tmpmapping->protocol, protocol))
     {
       mapping = tmpmapping;
@@ -871,7 +914,7 @@ gupnp_simple_igd_remove_port_real (GUPnPSimpleIgd *self,
               "DeletePortMapping",
               _service_proxy_delete_port_mapping, self,
               "NewRemoteHost", G_TYPE_STRING, "",
-              "NewExternalPort", G_TYPE_UINT, mapping->external_port,
+              "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
               "NewProtocol", G_TYPE_STRING, mapping->protocol,
               NULL);
 
